#include <stdio.h>
#include <math.h>
#include <string.h>

#include "iir.h"
#include "vec.h"

/* 
 * Apply a IIR filter defined by coeffs. on a & b (size lenCoefs). Data is on 
 * vec and filtered data goes to filtered (both size len).
 */
void recursive( signed short * vec, float * filtered, long len, float * a, float * b, int lenCoefs )
{
    long i, j;
    float tmp;

    for( i= 2; i<len; i++ )
    {
        filtered[i]= 0.0;


        for( j=0; j<lenCoefs; j++ )    
        {
            tmp= filtered[i];
            filtered[i]+= a[j] * (float)vec[i-j];
        }    
        for( j=1; j<lenCoefs; j++ )    
        {
            tmp= filtered[i];
            filtered[i]+= b[j] * filtered[i-j];
        }    
    }

}

/* 
 * Version for narrow pass-band iir filter. Like "recursive" but using three
 * coeffs. from a & b.
 */
void recursiveNarrow( signed short * vec, float * filtered, long len, float * a, float * b )
{
    long i;

    for( i= 2; i<len; i++ )
    {
        filtered[i]= a[0]*vec[i] + a[1]*vec[i-1] + a[2]*vec[i-2] 
            + b[1]*filtered[i-1] + b[2]*filtered[i-2];
    }
}

/*
 * Creates coeficients for a narrow pass-band IIR filter on a & b. Filter is
 * defined by center (band's center frequency) and bw (band size)
 * center and bw takes values between 0 & 0.5 (frequency range from 0 to 1/2 
 * of sampling frequency)
 */
void iirPassband( float * a, float * b, float center, float bw )
{
    float R, K;

    R= 1 - 3*bw;
    K= (1 - 2*R*cos(2*M_PI*center) + R*R)/(2 - 2*cos(2*M_PI*center));

    a[0]= 1 - K;
    a[1]= 2*(K - R) * cos(2*M_PI*center);
    a[2]= R*R - K;
    b[1]= 2*R*cos(2*M_PI*center);
    b[2]= -R*R;
}

/*
 * Creates a Chebyshev IIR filter and stores coefficients on a & b.
 * frec is the cutoff frequency
 * type: 
 *      0 -> lowpass
 *      1 -> hiwpass
 * ripple: filter's ripple 
 * poles: number of poles
 */
void chebyshev( float * a, float * b, float frec, int type, float ripple, int poles )
{
    int p, i;
    float aCoefs[3], bCoefs[3];
    float tmpA[22], tmpB[22];
    float sa, sb, gain;

    /*printf("2^3= %f\n", mipow(2, 3));
    printf("-1^0= %f\n", mipow(-1, 0));
    printf("-1^2= %f\n", mipow(-1, 2));*/

    zero( a, 22 );
    zero( b, 22 );

    a[2]= 1;
    b[2]= 1;

    for( p=1; p<=poles/2; p++ )
    {
        chebyshevCoefs( frec, type, ripple, poles, p, aCoefs, bCoefs );
        memcpy( tmpA, a, 22*sizeof(float) );
        memcpy( tmpB, b, 22*sizeof(float) );

        for( i=2; i<22; i++ )
        {
            a[i]= aCoefs[0]*tmpA[i] + aCoefs[1]*tmpA[i-1] + aCoefs[2]*tmpA[i-2];
            b[i]= tmpB[i] - bCoefs[1]*tmpB[i-1] - bCoefs[2]*tmpB[i-2];

            /*printf("1: a[%d]= %f\n", i, a[i]);
            printf("1: b[%d]= %f\n", i, b[i]);*/
        }
    }

    b[2]= 0;

    for( i=0; i<=20; i++ )
    {
        a[i]= a[i+2];
        b[i]= -b[i+2];
    }

    sa= 0;
    sb= 0;

    for( i=0; i<=20; i++ )
    {
        if ( type == 0 )
        {
            sa+= a[i];
            sb+= b[i];
        }
        else
        {
            sa+= a[i] * pow( -1, (double)i );
            sb+= b[i] * pow( -1, (double)i );
        }
    }
    /*printf("2: sa= %f\n", sa);
    printf("2: sb= %f\n", sb);*/

    gain= sa / (1-sb);

    for( i=0; i<=20; i++ )
    {
        a[i]/= gain;
        /*printf("3: a[%d]= %f\n", i, a[i]);*/
    }
}

/*
 * Function called by chebyshev
 * Used to create coefficients (aCoefs & bCoefs) needed for this func.
 */
void chebyshevCoefs( float frec, int type, float ripple, int nPoles, int pole, float * aCoefs, float * bCoefs )
{
    float rp, ip, es, vx, kx, k, t, w, m, d, x0, x1, x2, y1, y2;
    /*float M_PI= 3.141592;*/

    rp= -cos(M_PI/(nPoles*2) + (pole-1) * (M_PI/nPoles));
    ip=  sin(M_PI/(nPoles*2) + (pole-1) * (M_PI/nPoles));


    if ( ripple != 0 ) 
    {
        es= sqrt( pow( 100/(100-ripple), 2 ) -1 );
        vx= ((float)1/nPoles) * log( (1/es) + sqrt( (1/(es*es)) + 1) );
        kx= ((float)1/nPoles) * log( (1/es) + sqrt( (1/(es*es)) - 1) );
        kx= (exp(kx) + exp(-kx))/2;
        rp= rp * ( (exp(vx) - exp(-vx) ) /2 ) / kx;
        ip= ip * ( (exp(vx) + exp(-vx) ) /2 ) / kx;
    }

    /*printf( "rp= %f\nip= %f\nes= %f\nvx= %f\nkx= %f\n", rp, ip, es, vx, kx );*/

    t= 2 * tan(0.5);
    w= 2 * M_PI * frec;
    m= rp*rp + ip*ip;
    d= 4 - 4*rp*t + m*t*t;
    x0= (t*t)/d;
    x1= (2*t*t)/d;
    x2 = (t*t)/d;
    y1 = (8 - 2*m*t*t)/d;
    y2 = (-4 - 4*rp*t - m*t*t)/d;

    /*printf( "\nt= %f\nw= %f\nm= %f\nd= %f\nx0= %f\nx1= %f\nx2= %f\ny1= %f\ny2= %f\n\n", t, w, m, d, x0, x1, x2, y1, y2);*/

    if ( type == 1 )
        k = -cos(w/2 + 0.5) / cos(w/2 - 0.5);
    else    
        k =  sin(0.5 - w/2) / sin(0.5 + w/2);
        
    d = 1 + y1*k - y2*k*k;
    aCoefs[0] = (x0 - x1*k + x2*k*k)/d;
    aCoefs[1] = (-2*x0*k + x1 + x1*k*k - 2*x2*k)/d;
    aCoefs[2] = (x0*k*k - x1*k + x2)/d;
    bCoefs[1] = (2*k + y1 + y1*k*k - 2*y2*k)/d;
    bCoefs[2] = (-(k*k) - y1*k + y2)/d;

    if ( type == 1 )
    {
        aCoefs[1]= -aCoefs[1];
        bCoefs[1]= -bCoefs[1];
    }    
}
